"""Simulation Campaign configuration."""

import logging
from collections.abc import Iterator
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Optional

import numpy as np
import pandas as pd
import xarray as xr

from blueetl.constants import SIMULATION_PATH
from blueetl.types import StrOrPath
from blueetl.utils import dump_yaml, load_yaml, resolve_path

L = logging.getLogger(__name__)

SONATA_CONFIG = "simulation_config.json"
BLUEPY_CONFIG = "BlueConfig"


def _resolve_simulation_path(path_prefix: str, path: str, filename: str) -> str:
    """Convert to full absolute path, without checking if it exists."""
    if not path or path.startswith("https://"):
        # do not convert excluded simulations or nexus urls
        return path
    path_obj = resolve_path(path_prefix, path)
    if path_obj.name != filename:
        # the paths in the xarray simulation campaign config don't include the filename
        path_obj = path_obj / filename
    return str(path_obj)


def _reduce_simulation_path(path_prefix: str, path: str, filename: str) -> str:
    """Reduce to shorter path, without path_prefix and filename."""
    if not path or path.startswith("https://"):
        # do not convert excluded simulations or nexus urls
        return path
    path_obj = Path(path)
    if path_obj.is_relative_to(path_prefix):
        path_obj = path_obj.relative_to(path_prefix)
    if path_obj.name == filename:
        # the paths in the xarray simulation campaign config don't include the filename
        path_obj = path_obj.parent
    return str(path_obj)


@dataclass
class SimulationRow:
    """Simulation row in the simulation campaign."""

    index: int
    path: str
    conditions: dict[str, Any]

    @property
    def empty(self) -> bool:
        """Return True if the simulation has been excluded when running the campaign."""
        return not self.path


class SimulationCampaign:
    """Simulation campaign configuration."""

    def __init__(
        self,
        data: pd.DataFrame,
        name: str,
        attrs: dict,
        config_dir: Optional[Path] = None,
    ) -> None:
        """Init the configuration.

        Args:
            data: DataFrame of simulation paths, having as columns: simulation_path,
                and the parameters used for each simulation (for example: seed, ca...).
            name: name of the Simulation Campaign.
            attrs: dict of custom attributes.
            config_dir: if specified, it's used to resolve relative paths in attrs.
        """
        self._name = name
        self._attrs = attrs.copy()
        if config_dir:
            if "path_prefix" in self._attrs:
                self._attrs["path_prefix"] = str(
                    resolve_path(config_dir, self._attrs["path_prefix"])
                )
            if "circuit_config" in self._attrs:
                self._attrs["circuit_config"] = str(
                    resolve_path(config_dir, self._attrs["circuit_config"])
                )
        self._data = data.copy()
        self._data[SIMULATION_PATH] = self._resolve_paths(self._data[SIMULATION_PATH])

    def _get_simulation_filename(self) -> str:
        """Return the filename of each simulation in the campaign."""
        return SONATA_CONFIG if self.is_sonata() else BLUEPY_CONFIG

    def _resolve_paths(self, simulation_paths: pd.Series) -> pd.Series:
        """Resolve the simulation paths."""
        return simulation_paths.apply(
            lambda path: _resolve_simulation_path(
                path_prefix=self.attrs.get("path_prefix", ""),
                path=path,
                filename=self._get_simulation_filename(),
            )
        )

    def _reduce_paths(self, simulation_paths: pd.Series) -> pd.Series:
        """Reduce the simulation paths."""
        return simulation_paths.apply(
            lambda path: _reduce_simulation_path(
                path_prefix=self.attrs.get("path_prefix", ""),
                path=path,
                filename=self._get_simulation_filename(),
            )
        )

    def __eq__(self, other: object) -> bool:
        """Return True if the objects are considered equal, False otherwise."""
        if not isinstance(other, SimulationCampaign):
            # https://docs.python.org/3/library/constants.html#NotImplemented
            return NotImplemented
        return all(
            (
                self.name == other.name,
                self.attrs == other.attrs,
                self._data.equals(other._data),
            )
        )

    @property
    def name(self) -> str:
        """Return the name of the simulations campaign."""
        return self._name

    @property
    def attrs(self) -> dict:
        """Return the attributes dict associated with the simulations campaign."""
        return self._attrs

    @property
    def condition_names(self) -> list[str]:
        """Return the names of the parameters used to run the simulations."""
        return [c for c in self._data.columns if c != SIMULATION_PATH]

    @property
    def conditions(self) -> pd.DataFrame:
        """Return the DataFrame of the parameters used to run the simulations."""
        return self._data[self.condition_names]

    def is_coupled(self):
        """Return True if the coords are coupled, False otherwise."""
        return bool(self.attrs.get("__coupled__"))

    def is_sonata(self):
        """Return True if the simulations are in SONATA format, False otherwise."""
        circuit_config = self.attrs.get("circuit_config", "")
        if not circuit_config:
            raise RuntimeError("circuit_config is missing in the simulation campaign")
        return circuit_config.endswith(".json")

    @classmethod
    def load(cls, path: StrOrPath) -> "SimulationCampaign":
        """Load the configuration from file.

        Different file formats are supported:

        - xarray (json or yaml): configuration produced by bbp-workflow.
        - blueetl (json or yaml): configuration produced by this same class.

        Args:
            path: path to the configuration file.

        Returns:
            SimulationCampaign: new instance.
        """
        config = load_yaml(path)
        if config.get("format") == "blueetl":
            L.debug("Detected blueetl configuration format")
            return cls.from_dict(config, config_dir=Path(path).parent)
        if set(config) == {"name", "attrs", "data", "dims", "coords"}:
            L.debug("Detected xarray configuration format")
            return cls.from_xarray_dict(config, config_dir=Path(path).parent)
        raise ValueError("Unable to detect the configuration format")

    def dump(self, path: StrOrPath) -> None:
        """Save the configuration to file."""
        dump_yaml(path, data=self.to_dict())

    @classmethod
    def from_dict(cls, d: dict, config_dir: Optional[Path] = None) -> "SimulationCampaign":
        """Load the configuration from dict."""
        data = pd.DataFrame.from_records(d["data"])
        return cls(data=data, name=d["name"], attrs=d["attrs"], config_dir=config_dir)

    def to_dict(self) -> dict:
        """Convert the configuration to dict."""
        return {
            "format": "blueetl",
            "version": 1,
            "name": self.name,
            "attrs": self.attrs,
            "data": self._data.to_dict(orient="records"),
        }

    @classmethod
    def from_xarray(
        cls, da: xr.DataArray, config_dir: Optional[Path] = None
    ) -> "SimulationCampaign":
        """Load the configuration from xarray.DataArray."""
        df = da.to_dataframe(SIMULATION_PATH)
        # If the campaign is not coupled, the result after calling `to_dataframe()` is like:
        #
        #                    simulation_path
        # ca  depolarization
        # 1.0 3.0                     uuid/0
        #     4.0
        # 2.0 3.0                     uuid/2
        #     4.0                     uuid/3
        #
        # Otherwise, the result is like:
        #
        #           ca  depolarization simulation_path
        # coupled
        # 0        1.0             3.0          uuid/0
        # 1        2.0             4.0
        attrs = da.attrs.copy()
        if len(df.columns) == 1:
            coupled = None
        else:
            coupled = df.index.name
            assert coupled
            assert isinstance(df.index, pd.RangeIndex)
            # save the __coupled__ attribute to be able to convert back to xarray
            attrs["__coupled__"] = coupled
        data = df.reset_index(drop=bool(coupled))
        return cls(data=data, name=str(da.name), attrs=attrs, config_dir=config_dir)

    def to_xarray(self) -> xr.DataArray:
        """Convert the configuration to xarray.DataArray."""
        s = self._data.set_index(self.condition_names)[SIMULATION_PATH]
        s = self._reduce_paths(s)
        s.name = self.name
        attrs = self.attrs.copy()
        coupled = attrs.pop("__coupled__", None)
        if not coupled:
            # generated by GenerateSimulationCampaign
            da = xr.DataArray.from_series(s)
            da.attrs = attrs
        else:
            # generated by GenerateCoupledCoordsSimulationCampaign
            da = xr.DataArray(
                list(s),
                name=s.name,
                dims=coupled,
                coords={coupled: s.index},
                attrs=attrs,
            ).reset_index(coupled)
        return da

    @classmethod
    def from_xarray_dict(cls, d: dict, config_dir: Optional[Path] = None) -> "SimulationCampaign":
        """Load the configuration from a dict representing xarray.DataArray."""
        da = xr.DataArray.from_dict(d)
        return cls.from_xarray(da, config_dir=config_dir)

    def to_xarray_dict(self) -> dict:
        """Return the configuration as a dict representing xarray.DataArray."""
        return self.to_xarray().to_dict()

    def __len__(self) -> int:
        """Return the number of simulations."""
        return len(self._data)

    def __iter__(self) -> Iterator[SimulationRow]:
        """Iterate over the simulation rows."""
        for i, (_, sim_dict) in enumerate(self._data.etl.iterdict()):
            path = sim_dict.pop(SIMULATION_PATH)
            yield SimulationRow(index=i, path=path, conditions=sim_dict)

    def __getitem__(self, index: int) -> SimulationRow:
        """Return a specific simulation row."""
        sim_dict = self._data.loc[index].to_dict()
        path = sim_dict.pop(SIMULATION_PATH)
        return SimulationRow(index=index, path=path, conditions=sim_dict)

    def get(self, *args, **kwargs) -> pd.DataFrame:
        """Return a DataFrame with the selected simulations.

        See ``etl.q`` for the filter syntax.
        """
        return self._data.copy().etl.q(*args, **kwargs)

    def ids(self, *args, **kwargs) -> np.ndarray:
        """Return a numpy array with the ids of the selected simulations.

        See ``etl.q`` for the filter syntax.
        """
        return self.get(*args, **kwargs).index.to_numpy()



"""gid features from BlueCV."""

import logging
from functools import partial

import numpy as np
from elephant import statistics

from blueetl.constants import TIME

L = logging.getLogger(__name__)


def calculate_features_by_gid(repo, key, df, params):
    """Calculate features grouped by gid."""
    t_start, t_stop = repo.windows.get_bounds(key.window)
    spiketrain = df[TIME].to_numpy()
    functions = {
        "MFR": partial(get_MFR, spiketrain, t_start=t_start, t_stop=t_stop),
        "ISI": partial(get_ISI, spiketrain),
        "CV": partial(get_CV, spiketrain),
        "LV": partial(get_LV, spiketrain),
        "latency": partial(get_latency, spiketrain, t_start=t_start),
        "spike_count": partial(get_spike_count, spiketrain),
    }
    result = {}
    for feature_name, feature_config in params.items():
        feature_params = feature_config.get("params", {})
        result[feature_name] = functions[feature_name](**feature_params)
    return result


def get_MFR(spiketrain, t_start, t_stop):
    """Get mean firing rate of a neuron."""
    return statistics.mean_firing_rate(spiketrain, t_start=t_start, t_stop=t_stop)


def get_ISI(spiketrain):
    """Get interspike intervals of a neuron."""
    isi = statistics.isi(spiketrain)
    return isi[0] if isi.size > 0 else np.nan


def get_CV(spiketrain):
    """Get coefficient of variation of a neuron."""
    return statistics.cv(statistics.isi(spiketrain))


def get_LV(spiketrain):
    """Get measure of local variation for sequence of time intervals between events of a neuron."""
    return statistics.lv(statistics.isi(spiketrain), with_nan=True)


def get_latency(spiketrain, t_start):
    """Get latency of a single neuron (time until first spike)."""
    return spiketrain[0] - t_start


def get_spike_count(spiketrain):
    """Get number of spikes in a time window per gid."""
    return spiketrain.size



"""NeuronClasses extractor."""

import json
import logging

import pandas as pd

from blueetl.config.analysis_model import NeuronClassConfig
from blueetl.constants import (
    CIRCUIT_ID,
    COUNT,
    GIDS,
    LIMIT,
    NEURON_CLASS,
    NODE_SET,
    POPULATION,
    QUERY,
)
from blueetl.extract.base import BaseExtractor
from blueetl.extract.neurons import Neurons

L = logging.getLogger(__name__)


class NeuronClasses(BaseExtractor):
    """NeuronClasses extractor class."""

    COLUMNS = [
        CIRCUIT_ID,
        NEURON_CLASS,
        COUNT,
        LIMIT,
        POPULATION,
        NODE_SET,
        GIDS,
        QUERY,
    ]

    @classmethod
    def from_neurons(
        cls, neurons: Neurons, neuron_classes: dict[str, NeuronClassConfig]
    ) -> "NeuronClasses":
        """Load neuron classes information for each circuit.

        Args:
            neurons: Neurons extractor.
            neuron_classes: configuration dict of neuron classes to be extracted.

        Returns:
            NeuronClasses: new instance.
        """
        results = []
        for index, count in neurons.count_by_neuron_class().etl.iter():
            # index: circuit_id, neuron_class
            config = neuron_classes[index.neuron_class]
            # gids and query are saved for reference and inspection.
            # query is saved as json string to avoid this error during the conversion to parquet:
            # pyarrow.lib.ArrowInvalid: 'cannot mix list and non-list, non-null values'
            results.append(
                {
                    CIRCUIT_ID: index.circuit_id,
                    NEURON_CLASS: index.neuron_class,
                    COUNT: count,
                    LIMIT: config.limit,
                    POPULATION: config.population,
                    NODE_SET: config.node_set,
                    GIDS: json.dumps(config.node_id) if config.node_id is not None else None,
                    QUERY: json.dumps(config.query),
                }
            )
        df = pd.DataFrame(results)
        return cls(df, cached=False, filtered=False)



"""Simulations extractor."""

import logging
from enum import Enum
from pathlib import Path
from typing import Any, Optional, cast

import pandas as pd

from blueetl.adapters.circuit import CircuitAdapter as Circuit
from blueetl.adapters.simulation import SimulationAdapter as Simulation
from blueetl.campaign.config import SimulationCampaign
from blueetl.constants import CIRCUIT, CIRCUIT_ID, SIMULATION, SIMULATION_ID, SIMULATION_PATH
from blueetl.extract.base import BaseExtractor

L = logging.getLogger(__name__)

# temporary column in the simulations dataframe
_STATUS = "_status"


class SimulationStatus(Enum):
    """Status of the loaded simulation."""

    COMPLETE = "COMPLETE"
    INCOMPLETE = "INCOMPLETE"
    MISSING = "MISSING"


class InconsistentSimulations(Exception):
    """Error raised when the extracted simulations have some inconsistencies."""


class Simulations(BaseExtractor):
    """Simulations extractor class."""

    COLUMNS = [SIMULATION_PATH, SIMULATION_ID, CIRCUIT_ID, SIMULATION, CIRCUIT]
    # allow additional columns containing the simulation conditions
    _allow_extra_columns = True

    @classmethod
    def _build_record(
        cls,
        simulation_id: int,
        rec: dict[str, Any],
        circuit_hashes: dict[str, int],
        circuits: dict[int, Circuit],
    ) -> dict[str, Any]:
        """Build a record to be added to the simulations dataframe."""
        # use the cached simulation_id if available
        simulation_id = rec.get(SIMULATION_ID, simulation_id)
        # use the cached circuit_id if available, or fallback to simulation_id
        circuit_id = rec.get(CIRCUIT_ID, simulation_id)
        simulation_path = rec[SIMULATION_PATH]
        # use the same Simulation and Circuit objects if available
        simulation = cast(Optional[Simulation], rec.get(SIMULATION))
        circuit = cast(Optional[Circuit], rec.get(CIRCUIT))
        assert (
            simulation and circuit or not simulation and not circuit
        ), "Simulation and Circuit must be both initialized, or both not initialized"
        circuit_hash = None
        status = SimulationStatus.MISSING
        simulation = simulation or Simulation.from_file(Path(simulation_path))
        if simulation.exists():
            # consider the simulation only if it wasn't manually deleted
            status = SimulationStatus.INCOMPLETE
            circuit_hash = simulation.circuit.checksum()
            # if circuit_hash is not new, use the previous circuit_id
            circuit_id = circuit_hashes.setdefault(circuit_hash, circuit_id)
            circuit = circuit or circuits.setdefault(circuit_id, simulation.circuit)
            # double-check the circuit config hash in case it was cached
            if circuit_hash != circuit.checksum():
                L.error("Inconsistent circuit hash and id, you may need to delete the cache")
                raise InconsistentSimulations("Inconsistent hash and id")
            if simulation.is_complete():
                status = SimulationStatus.COMPLETE
        sim_repr = f"{simulation_id=}, {circuit_id=}, {circuit_hash=}, {simulation_path=}"
        L.debug("Processing simulation: %s", sim_repr)
        return {
            SIMULATION_ID: simulation_id,
            CIRCUIT_ID: circuit_id,
            SIMULATION_PATH: simulation_path,
            SIMULATION: simulation,
            CIRCUIT: circuit,
            _STATUS: status,
        }

    @classmethod
    def _from_paths(cls, simulation_paths: pd.DataFrame) -> pd.DataFrame:
        """Return a dataframe of simulations from a list of simulation paths.

        Args:
            simulation_paths: DataFrame containing the simulation paths.
                Any additional columns besides (simulation_path, simulation_id, circuit_id)
                are returned unchanged in the resulting DataFrame.

        Returns:
            DataFrame of simulations.
        """
        circuit_hashes: dict[str, int] = {}  # map circuit_hash -> circuit_id
        circuits: dict[int, Circuit] = {}  # map circuit_id -> circuit
        records = []
        for simulation_id, (_, rec) in enumerate(simulation_paths.etl.iterdict()):
            record = cls._build_record(
                simulation_id=simulation_id,
                rec=rec,
                circuit_hashes=circuit_hashes,
                circuits=circuits,
            )
            # rec must be first for lower precedence in case of collisions
            records.append({**rec, **record})
        assert len(set(circuit_hashes.values())) == len(circuit_hashes), "Inconsistent circuit ids"
        return pd.DataFrame(records)

    @classmethod
    def _filter_simulations_df(
        cls, df: pd.DataFrame, query: Optional[dict], cached: bool = False
    ) -> pd.DataFrame:
        """Remove the missing and incomplete simulations and filter by the given query.

        Args:
            df: DataFrame to filter.
            query: filtering query.
            cached: if True, raise an error in case any simulation is missing or incomplete.

        Returns:
            the filtered DataFrame.
        """
        len_total = len(df)
        len_missing = len(df.etl.q({_STATUS: SimulationStatus.MISSING}))
        len_incomplete = len(df.etl.q({_STATUS: SimulationStatus.INCOMPLETE}))
        # filter by the internal ``_status`` column, and drop the column
        df = df.etl.q({_STATUS: SimulationStatus.COMPLETE}).drop(columns=_STATUS)
        len_complete = len(df)
        assert len_missing + len_incomplete + len_complete == len_total
        # filter by the custom query if provided
        df = df.etl.q(query or {})
        # reset the index to ensure that it doesn't contain gaps,
        # while preserving the simulation_id in a dedicated column
        df = df.reset_index(drop=True)
        len_final = len(df)
        L.info("Simulations ignored because missing: %s", len_missing)
        L.info("Simulations ignored because incomplete: %s", len_incomplete)
        L.info("Simulations filtered out: %s, with query: %s", len_complete - len_final, query)
        L.info(
            "Simulations extracted: %s/%s, ids: %s",
            len_final,
            len_total,
            df[SIMULATION_ID].to_list(),
        )
        if cached and (len_missing or len_incomplete):
            L.error(
                "Some cached simulations are missing or incomplete, "
                "you may need to delete the cache."
            )
            raise InconsistentSimulations("Inconsistent cache")
        return df

    @classmethod
    def from_config(cls, config: SimulationCampaign, query: Optional[dict] = None) -> "Simulations":
        """Extract simulations from the given simulation campaign."""
        df = config.get()
        return cls.from_pandas(df=df, query=query, cached=False)

    @classmethod
    def from_pandas(
        cls,
        df: pd.DataFrame,
        query: Optional[dict] = None,
        cached: bool = True,
    ) -> "Simulations":
        """Extract simulations from a dataframe containing valid simulation ids and circuit ids."""
        original_len = len(df)
        df = cls._from_paths(df)
        df = cls._filter_simulations_df(df, query, cached=cached)
        filtered = len(df) != original_len
        return cls(df, cached=cached, filtered=filtered)

    def to_pandas(self) -> pd.DataFrame:
        """Dump simulations to a dataframe that can be serialized and stored."""
        # skip columns SIMULATION and CIRCUIT because they contain custom runtime objects
        columns = [col for col in self.df.columns if col not in [SIMULATION, CIRCUIT]]
        return self.df[columns]



import json
from unittest.mock import MagicMock

import pandas as pd
import pytest
from pandas.testing import assert_frame_equal

from blueetl.config.analysis_model import NeuronClassConfig
from blueetl.constants import CIRCUIT_ID, COUNT, GID, NEURON_CLASS
from blueetl.extract.neuron_classes import NeuronClasses
from blueetl.utils import ensure_dtypes


def _count_by_neuron_class(data):
    """Return a Series as returned by neurons.count_by_neuron_class()."""
    df = pd.DataFrame(data, columns=[CIRCUIT_ID, NEURON_CLASS, GID])
    return df.set_index([CIRCUIT_ID, NEURON_CLASS])[GID]


def test_neuron_classes_from_neurons():
    mock_neurons = MagicMock()
    mock_neurons.count_by_neuron_class.return_value = _count_by_neuron_class(
        [
            [0, "L23_EXC", 123],
            [0, "L4_INH", 456],
        ]
    )
    result = NeuronClasses.from_neurons(
        mock_neurons,
        neuron_classes={
            "L23_EXC": NeuronClassConfig(
                **{
                    "population": "thalamus_neurons",
                    "query": {"layer": ["2", "3"], "synapse_class": ["EXC"]},
                }
            ),
            "L4_INH": NeuronClassConfig(
                **{
                    "population": "thalamus_neurons",
                    "query": {"layer": ["4"], "synapse_class": ["INH"]},
                }
            ),
        },
    )
    assert mock_neurons.count_by_neuron_class.call_count == 1
    assert isinstance(result, NeuronClasses)
    expected_df = pd.DataFrame(
        [
            {
                CIRCUIT_ID: 0,
                NEURON_CLASS: "L23_EXC",
                COUNT: 123,
                "limit": None,
                "population": "thalamus_neurons",
                "node_set": None,
                "gids": None,
                "query": json.dumps({"layer": ["2", "3"], "synapse_class": ["EXC"]}),
            },
            {
                CIRCUIT_ID: 0,
                NEURON_CLASS: "L4_INH",
                COUNT: 456,
                "limit": None,
                "population": "thalamus_neurons",
                "node_set": None,
                "gids": None,
                "query": json.dumps({"layer": ["4"], "synapse_class": ["INH"]}),
            },
        ]
    )
    expected_df = ensure_dtypes(expected_df)
    assert_frame_equal(result.df, expected_df)


def test_neuron_classes_from_neurons_without_neurons():
    mock_neurons = MagicMock()
    mock_neurons.count_by_neuron_class.return_value = _count_by_neuron_class([])
    with pytest.raises(RuntimeError, match="No data extracted to NeuronClasses"):
        NeuronClasses.from_neurons(
            mock_neurons,
            neuron_classes={
                "aclass": NeuronClassConfig(
                    **{"population": "thalamus_neurons", "query": {"region": "any"}}
                )
            },
        )
    assert mock_neurons.count_by_neuron_class.call_count == 1



"""Parallelization utilities."""

import logging
from collections import namedtuple
from collections.abc import Callable, Iterator
from functools import partial
from types import SimpleNamespace
from typing import Any, NamedTuple

import pandas as pd
from blueetl_core.parallel import Task, run_parallel
from blueetl_core.utils import CachedDataFrame

from blueetl.constants import CIRCUIT_ID, SIMULATION_ID

L = logging.getLogger(__name__)


def _unique_rows(df: pd.DataFrame, columns: list[str]) -> pd.DataFrame:
    """Return the unique rows of the given DataFrame.

    - Only the specified columns are used to determine the uniqueness of the rows.
    - If a column is not present in the DataFrame, it's ignored.
    """
    if any(df.index.names):
        # reset the index if any columns are contained in the index levels
        if levels := [level for level in columns if level in df.index.names]:
            df = df.reset_index(level=levels)
    # find the unique rows
    df = df[df.columns.intersection(columns)].drop_duplicates(ignore_index=True)
    L.info("Considering %s rows for columns %s", len(df), list(df.columns))
    return df


def _groups(df_list: list[pd.DataFrame], groupby: list[str]) -> pd.DataFrame:
    """Get all the possible groups and return a sorted merged dataframe.

    The result is similar to what you get joining all the DataFrames and using groupby
    on the result to get all the groups, but it should use less memory and may be faster
    with big DataFrames.

    Dataframes not containing any column or index level in groupby are ignored.
    """
    result = None
    for df in df_list:
        df = _unique_rows(df, groupby)
        if len(df) > 0:
            if result is None:
                result = df
            else:
                result = result.merge(df, how="left")
            if set(result.columns) == set(groupby):
                # all the groupby columns have already been considered
                break
    if result is None:
        raise RuntimeError("At least one filtering DataFrame must be specified")
    if diff := set(groupby) - set(result.columns):
        raise ValueError(f"Unknown columns in groupby: {sorted(diff)}")
    # ensure that the groups are sorted by the specified columns
    return result[groupby].sort_values(groupby, ignore_index=True)


def _func_generator(
    df_list: list[pd.DataFrame], groupby: list[str], func: Callable
) -> Iterator[Callable]:
    """Yield functions to be executed in a subprocess."""
    groups = _groups(df_list, groupby=groupby)
    caches = [CachedDataFrame(df) for df in df_list]
    L.info("Tasks to be executed: %s", len(groups))
    # for each group, yield a function that can be called in a subprocess
    for _, key in groups.etl.iter():
        filtered = [df.query(key._asdict(), ignore_unknown_keys=True) for df in caches]
        yield partial(func, key=key, df_list=filtered)


def merge_filter(
    df_list: list[pd.DataFrame],
    groupby: list[str],
    func: Callable[[int, NamedTuple, list[pd.DataFrame]], Any],
) -> list:
    """Merge the specified columns of the list of DataFrames, and call func for each combination.

    The merge operation is similar to a SQL left outer join.

    Args:
        df_list: list of DataFrames.
        groupby: list of columns to consider across the DataFrames.
        func: callback executed for each calculated combination of columns, with parameters:

            - task_index (int): task index.
            - key (NamedTuple): key used to filter the DataFrames passed to each function call.
            - df_list (list[pd.DataFrames]): list of DataFrames filtered by key.

    Returns:
        list of values returned by the callback function.

    """
    func_generator = _func_generator(df_list=df_list, groupby=groupby, func=func)
    task_generator = (Task(partial(f, task_index=i)) for i, f in enumerate(func_generator))
    return run_parallel(task_generator)


def merge_groupby(
    df_list: list[pd.DataFrame], groupby: list[str]
) -> Iterator[tuple[NamedTuple, pd.DataFrame]]:
    """Merge a list of DataFrames, group by the given keys, and yield keys and groups.

    The merge operation is similar to a SQL left outer join, but
    the dataframes are filtered in the main process and merged in subprocesses.
    """

    def _func(
        task_index: int, key: NamedTuple, df_list: list[pd.DataFrame]
    ) -> tuple[NamedTuple, pd.DataFrame]:
        # pylint: disable=unused-argument
        # executed in a subprocess
        merged = df_list[0]
        for df in df_list[1:]:
            merged = merged.merge(df, how="left", copy=False)
        return key, merged

    yield from merge_filter(df_list=df_list, groupby=groupby, func=_func)


def call_by_simulation(
    simulations: pd.DataFrame,
    dataframes_to_filter: dict[str, pd.DataFrame],
    func: Callable,
    how: str = "dataframe",
) -> list[Any]:
    """Execute the given function in parallel, one task for each simulation.

    Args:
        simulations: DataFrame of simulations.
        dataframes_to_filter: dict of DataFrames to filter by simulation_id and/or circuit_id,
            and passed to each subprocess.
        func: callable called for each simulation, accepting:
            - simulation_row: NamedTuple (or the type specified with the `how` parameter)
            - filtered_dataframes: dict of DataFrames filtered by simulation_id and/or circuit_id
            If the function has other parameters, they can be applied using `functools.partials`,
            and they will be serialized and passed unchanged to the subprocesses.
        jobs: number of jobs (see run_parallel)
        backend: parallel backend (see run_parallel)
        how: format the `simulation_row` parameter passed to the func callback.
            It can be one of "namespace", "namedtuple", "dict", "series", "dataframe".

    Returns:
        list of results
    """

    def _func(
        task_index: int, key: NamedTuple, df_list: list[pd.DataFrame]
    ) -> tuple[NamedTuple, pd.DataFrame]:
        # pylint: disable=unused-argument
        # executed in a subprocess
        simulation_row = convert_row(df_list[0].reset_index())
        filtered_dataframes = dict(zip(dataframes_to_filter, df_list[1:]))
        return func(simulation_row=simulation_row, filtered_dataframes=filtered_dataframes)

    convert_row = {
        "namedtuple": lambda df: namedtuple("Values", df.columns)(**df.iloc[0]),
        "namespace": lambda df: SimpleNamespace(**df.iloc[0]),
        "dict": lambda df: df.iloc[0].to_dict(),
        "series": lambda df: df.iloc[0],
        "dataframe": lambda df: df,
    }[how]
    return list(
        merge_filter(
            df_list=[simulations, *dataframes_to_filter.values()],
            groupby=[SIMULATION_ID, CIRCUIT_ID],
            func=_func,
        )
    )



"""Windows extractor."""

import logging
from pathlib import Path
from typing import Any, Optional, Union

import numpy as np
import pandas as pd

from blueetl.adapters.circuit import CircuitAdapter as Circuit
from blueetl.adapters.node_sets import NodeSetsAdapter as NodeSets
from blueetl.adapters.simulation import SimulationAdapter as Simulation
from blueetl.config.analysis_model import TrialStepsConfig, WindowConfig
from blueetl.constants import (
    CHECKSUM_SEP,
    CIRCUIT_ID,
    DURATION,
    LEVEL_SEP,
    OFFSET,
    SIMULATION_ID,
    T_START,
    T_STEP,
    T_STOP,
    TRIAL,
    WINDOW,
    WINDOW_TYPE,
)
from blueetl.extract.base import BaseExtractor
from blueetl.extract.simulations import Simulations
from blueetl.resolver import Resolver
from blueetl.utils import import_by_string, timed

L = logging.getLogger(__name__)


def _load_dynamic_gids(
    circuit: Circuit,
    population: Optional[str],
    node_set: Optional[str],
    node_sets_file: Optional[Path],
    limit: Optional[int],
) -> np.ndarray:
    """Return the node ids to consider."""
    node_set = node_set or None
    with timed(L.info, "Loading nodes from circuit for dynamic offset"):
        if node_set and node_sets_file:
            node_sets = NodeSets.from_file(circuit.node_sets_file)
            node_sets |= NodeSets.from_file(node_sets_file)
            node_set = node_sets.instance[node_set]
        gids = circuit.nodes[population].ids(group=node_set)
    neuron_count = len(gids)
    if limit and neuron_count > limit:
        gids = np.random.choice(gids, size=limit, replace=False)
    L.info("Selected %s/%s gids", len(gids), neuron_count)
    return gids


def _load_dynamic_spikes(
    simulation: Simulation,
    population: Optional[str],
    gids: np.ndarray,
    offset: float,
    t_start: float,
    t_stop: float,
) -> np.ndarray:
    """Return the spikes extracted for the given parameters."""
    spikes = (
        simulation.spikes[population]
        .get(gids, t_start=offset + t_start, t_stop=offset + t_stop)
        .index.to_numpy()
    )
    spikes -= offset
    L.info("Selected %s spikes", len(spikes))
    return spikes


def _calculate_dynamic_offset(
    simulation: Simulation,
    circuit: Circuit,
    initial_offset: float,
    step_offsets: list[float],
    trial_steps_config: TrialStepsConfig,
) -> float:
    """Calculate the dynamic offset according to NSETM-2281."""
    # circuit is passed explicitly instead of loading it from simulation.circuit
    # to take advantage of any data already loaded and cached
    gids = _load_dynamic_gids(
        circuit=circuit,
        population=trial_steps_config.population,
        node_set=trial_steps_config.node_set,
        node_sets_file=trial_steps_config.node_sets_file,
        limit=trial_steps_config.limit,
    )
    spikes_list = []
    t_start, t_stop = trial_steps_config.bounds
    for step_offset in step_offsets:
        spikes = _load_dynamic_spikes(
            simulation=simulation,
            population=trial_steps_config.population,
            gids=gids,
            offset=initial_offset + step_offset,
            t_start=t_start,
            t_stop=t_stop,
        )
        spikes_list.append(spikes)
    func = import_by_string(trial_steps_config.function)
    result = func(spikes_list, trial_steps_config.dict())
    if not np.issubdtype(type(result), np.number):
        raise ValueError(f"The function {trial_steps_config.function} must return a number")
    return result


class Windows(BaseExtractor):
    """Windows extractor class."""

    COLUMNS = [
        SIMULATION_ID,
        CIRCUIT_ID,
        WINDOW,
        TRIAL,
        OFFSET,
        T_START,
        T_STOP,
        T_STEP,
        DURATION,
        WINDOW_TYPE,
    ]

    @classmethod
    def _validate(cls, df: pd.DataFrame) -> None:
        super()._validate(df)
        # check that all the trials in the same window have the same t_start, t_stop, duration
        if not np.all(
            df.groupby(WINDOW, observed=True)[[T_START, T_STOP, DURATION]].nunique() == 1
        ):
            raise ValueError("Inconsistent window t_start, t_stop, or duration in some trial(s)")
        # check that the trials numbers are unique and starting from 0
        if not np.all(
            df.groupby([SIMULATION_ID, WINDOW], observed=True)[TRIAL].agg(
                lambda x: sorted(x) == list(range(len(x)))
            )
        ):
            raise ValueError("Inconsistent trial index in some window(s)")

    @classmethod
    def _load_records_from_resolver(
        cls,
        name: str,
        win: str,
        simulation_id: int,
        circuit_id: int,
        resolver: Resolver,
    ) -> list[dict[str, Any]]:
        # example of valid win: spikes.extraction.windows.w1#checksum
        win, _, _checksum = win.rpartition(CHECKSUM_SEP)
        ref, _, window = win.rpartition(LEVEL_SEP)
        windows = resolver.get(ref)
        df = windows.df.etl.q(simulation_id=simulation_id, circuit_id=circuit_id, window=window)
        if len(df) == 0:
            raise RuntimeError(f"The window referenced by {win} is empty or undefined")
        return df.assign(**{WINDOW: name}).to_dict(orient="records")

    @classmethod
    def _load_records_from_config(
        cls,
        name: str,
        rec: Any,  # row from simulations DataFrame
        win: WindowConfig,
        trial_steps_config: Optional[TrialStepsConfig],
    ) -> list[dict[str, Any]]:
        """Load the records from the window configuration."""
        t_start, t_stop = win.bounds
        t_step = win.t_step
        duration = t_stop - t_start
        if win.trial_steps_list:
            step_offsets = win.trial_steps_list
        else:
            step_offsets = [win.trial_steps_value * i for i in range(win.n_trials or 1)]
        if trial_steps_config:
            dynamic_offset = _calculate_dynamic_offset(
                simulation=rec.simulation,
                circuit=rec.circuit,
                initial_offset=win.initial_offset,
                step_offsets=step_offsets,
                trial_steps_config=trial_steps_config,
            )
        else:
            dynamic_offset = 0.0
        L.info(
            "Using window=%s, initial_offset=%s, dynamic_offset=%s, step_offsets=%s, "
            "t_start=%s, t_stop=%s, t_step=%s, duration=%s",
            name,
            win.initial_offset,
            dynamic_offset,
            step_offsets,
            t_start,
            t_stop,
            t_step,
            duration,
        )
        return [
            {
                SIMULATION_ID: rec.simulation_id,
                CIRCUIT_ID: rec.circuit_id,
                WINDOW: name,
                TRIAL: index,
                OFFSET: win.initial_offset + dynamic_offset + step_offset,
                T_START: t_start,
                T_STOP: t_stop,
                T_STEP: t_step,
                DURATION: duration,
                WINDOW_TYPE: win.window_type,
            }
            for index, step_offset in enumerate(step_offsets)
        ]

    @classmethod
    def from_simulations(
        cls,
        simulations: Simulations,
        windows_config: dict[str, Union[str, WindowConfig]],
        trial_steps_config: dict[str, TrialStepsConfig],
        resolver: Resolver,
    ) -> "Windows":
        """Return a new Windows instance from the given simulations and configuration.

        Args:
            simulations: Simulations extractor.
            windows_config: configuration dict.
            trial_steps_config: configuration dict.
            resolver: resolver instance.

        Returns:
            Windows: new instance.
        """
        results = []
        for _, rec in simulations.df.etl.iter():
            for name, win in windows_config.items():
                L.info(
                    "Processing simulation_id=%s, circuit_id=%s, window=%s",
                    rec.simulation_id,
                    rec.circuit_id,
                    name,
                )
                if isinstance(win, str):
                    partial_results = cls._load_records_from_resolver(
                        name=name,
                        win=win,
                        simulation_id=rec.simulation_id,
                        circuit_id=rec.circuit_id,
                        resolver=resolver,
                    )
                else:
                    partial_results = cls._load_records_from_config(
                        name=name,
                        win=win,
                        rec=rec,
                        trial_steps_config=(
                            trial_steps_config[win.trial_steps_label]
                            if win.trial_steps_label
                            else None
                        ),
                    )
                results.extend(partial_results)

        df = pd.DataFrame(results)
        return cls(df, cached=False, filtered=False)

    def get_bounds(self, window: str) -> tuple[float, float]:
        """Return the interval (t_start, t_stop) for the specified window.

        The returned values don't depend on the simulation or the trial,
        because they are relative the offset, that is the only changing value.
        """
        rec = self.df.etl.first(window=window)
        t_start, t_stop = rec[[T_START, T_STOP]]
        return t_start, t_stop

    def get_duration(self, window: str) -> float:
        """Return the duration of the specified window."""
        return self.df.etl.first(window=window)[DURATION]

    def get_number_of_trials(self, window: str) -> int:
        """Return the number of trials for the specified window."""
        return np.max(self.df.etl.q(window=window)[TRIAL]) + 1



"""Resolver."""

from abc import ABC, abstractmethod
from typing import Any

from blueetl.constants import LEVEL_SEP


class ResolverError(Exception):
    """Error raised when the requested reference cannot be resolved."""


class Resolver(ABC):
    """Resolver class."""

    def __init__(self, root: Any) -> None:
        """Initialize the Resolver.

        Args:
            root: referenced root object.
        """
        self._root = root

    def get(self, ref: str) -> Any:
        """Return the object referred by the ref string.

        Args:
            ref: string with attributes separated by dot.
        """
        obj = self._root
        names = ref.split(LEVEL_SEP)
        for name in names:
            try:
                obj = self._get(obj, name)
            except Exception:
                msg = f"Impossible to resolve {LEVEL_SEP.join(names)} at level {name}"
                raise ResolverError(msg) from None
        return obj

    @abstractmethod
    def _get(self, obj: Any, name: str) -> Any:
        """Return the item or attr corresponding to the given name."""


class AttrResolver(Resolver):
    """AttrResolver class.

    Examples:
        >>> from unittest.mock import Mock
        >>> obj = Mock()  # any object with nested attributes
        >>> obj.a.b.c = 123
        >>> resolver = AttrResolver(obj)
        >>> resolver.get("a.b.c")
        123
    """

    def _get(self, obj: Any, name: str) -> Any:
        return getattr(obj, name)


class ItemResolver(Resolver):
    """ItemResolver class.

    Examples:
        >>> obj = {"a": {"b": {"c": 123}}}
        >>> resolver = ItemResolver(obj)
        >>> resolver.get("a.b.c")
        123
    """

    def _get(self, obj: Any, name: str) -> Any:
        return obj[name]



"""Bluepy implementation."""

from blueetl.utils import import_optional_dependency

# Immediately raise an error with a custom message if the submodule is imported
import_optional_dependency("bluepy")



"""Feather data store."""

import logging
from typing import Optional

import pandas as pd

from blueetl.store.base import BaseStore
from blueetl.utils import timed

L = logging.getLogger(__name__)
_idx_prefix = "_index"
_idx_sep = ":"


def _index_to_columns(df: pd.DataFrame) -> pd.DataFrame:
    """Convert index to columns in the given DataFrame.

    Feather does not support serializing generic Indexes and any MultiIndex for the index
    see https://github.com/pandas-dev/pandas/blob/v1.4.1/pandas/io/feather_format.py#L59-L81
    """
    if (
        not isinstance(df.index, (pd.Index, pd.RangeIndex))
        or df.index.name
        or not df.index.equals(pd.RangeIndex.from_range(range(len(df))))
    ):
        new_columns = {}
        for i, name in enumerate(df.index.names):
            # in case of unnamed MultiIndex, df.reset_index() automatically assigns level_N
            key = name or f"level_{i}"
            value = f"{_idx_prefix}{_idx_sep}{i}{_idx_sep}{name or ''}"
            new_columns[key] = value
        return df.reset_index().rename(columns=new_columns)
    return df


def _columns_to_index(df: pd.DataFrame) -> pd.DataFrame:
    """Convert columns to index in the given DataFrame."""
    mapping = {}
    for col in df.columns:
        if col.startswith(_idx_prefix):
            _, _, name = col.split(_idx_sep, 2)
            mapping[col] = name or None
    if mapping:
        df = df.set_index(list(mapping))
        # it works also for single level indexes
        df.index.set_names([mapping.get(name) for name in df.index.names], inplace=True)
    return df


class FeatherStore(BaseStore):
    """Feather data store."""

    @property
    def extension(self) -> str:
        """Return the file extension to be used with this specific data store."""
        return "feather"

    def dump(self, df: pd.DataFrame, name: str) -> None:
        """Save a dataframe to file, using the given name and the class extension."""
        path = self.path(name)
        with timed(L.debug, f"Writing {name} to {path}"):
            df = _index_to_columns(df)
            df.to_feather(path)

    def load(self, name: str) -> Optional[pd.DataFrame]:
        """Load a dataframe from file, using the given name and the class extension."""
        path = self.path(name)
        if not path.exists():
            return None
        with timed(L.debug, f"Reading {name} from {path}"):
            df = pd.read_feather(path)
            return _columns_to_index(df)



"""Parquet data store."""

import logging
from pathlib import Path
from typing import Any, Optional

import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq

from blueetl.store.base import BaseStore
from blueetl.types import StrOrPath
from blueetl.utils import timed

L = logging.getLogger(__name__)


def _get_unified_schema(path: StrOrPath) -> pa.Schema:
    """Infer and return the unified schema from all the parquet files in the given directory.

    Needed because pq.read_table() would infer the schema from just the first file,
    unless a metadata file is used. See:
    https://github.com/ueshin/apache-arrow/blob/0a2cf3ac/python/pyarrow/parquet.py#L724

    This can cause problems when a column in the first file contains empty lists,
    since they are considered as:

      times: list<element: null>
        child 0, element: null

    while in other files containing lists of floats they are considered as:

      times: list<element: double>
        child 0, element: double

    and with pyarrow 15.0.0 this would raise:

      ArrowNotImplementedError:
        Unsupported cast from double to null using function cast_null

    Processing all the parquet files could require some time (around 4 seconds for 13824 files
    stored in the shared memory /dev/shm), so it doesn't parse all the files if possible.

    If the performance are improved, we could replace the content of the function with:

        schemas = [pq.read_schema(file_path) for file_path in Path(path).iterdir()]
        return pa.unify_schemas(schemas)
    """
    with timed(L.debug, f"Inferring schema for {path}") as messages:
        schemas = []
        null_list_type = pa.list_(pa.null())
        valid_fields: dict[str, bool] = {}
        for file_path in Path(path).iterdir():
            all_valid = True
            file_schema = pq.read_schema(file_path)
            for field in file_schema:
                if valid_fields.get(field.name, False) is False:
                    valid_fields[field.name] = not field.type.equals(null_list_type)
                    all_valid = all_valid and valid_fields[field.name]
            schemas.append(file_schema)
            if all_valid:
                # break because the merged schema doesn't contain any list of nulls
                break
        messages.append(f"with {len(schemas)} loaded schemas")
        return pa.unify_schemas(schemas)


class ParquetStore(BaseStore):
    """Parquet data store."""

    def __init__(self, basedir: StrOrPath) -> None:
        """Initialize the object."""
        super().__init__(basedir=basedir)
        self._dump_options: dict[str, Any] = {
            "engine": "pyarrow",
            # zstd typically provides a higher compression ratio than snappy,
            # and the cpu time is similar
            "compression": "zstd",
        }
        self._load_options: dict[str, Any] = {
            "engine": "pyarrow",
        }

    @property
    def extension(self) -> str:
        """Return the file extension to be used with this specific data store."""
        return "parquet"

    def dump(self, df: pd.DataFrame, name: str, **kwargs) -> None:
        """Save a dataframe to file, using the given name and the class extension."""
        path = self.path(name)
        # Unless the parameter "index" is explicitly enforced, ensure that RangeIndex
        # is converted to Int64Index in MultiIndexes with Pandas >= 1.5.0.
        # See https://github.com/apache/arrow/issues/33030
        index = True if isinstance(df.index, pd.MultiIndex) else None
        with timed(L.debug, f"Writing {name or 'files'} to {path}"):
            df.to_parquet(path=path, **{"index": index, **self._dump_options, **kwargs})

    def load(self, name: str = "", **kwargs) -> Optional[pd.DataFrame]:
        """Load a dataframe from file, using the given name and the class extension.

        If name is empty, then consider and load all the files in the directory.
        """
        path = self.path(name)
        if not path.exists():
            return None
        if "schema" in kwargs:
            schema = kwargs.pop("schema")
        elif path.is_dir():
            schema = _get_unified_schema(path)
        else:
            schema = None
        with timed(L.debug, f"Reading {name or 'files'} from {path}"):
            return pd.read_parquet(path=path, **{"schema": schema, **self._load_options, **kwargs})



import json
from pathlib import Path

import pandas as pd
import pytest
from numpy.testing import assert_array_equal

from blueetl.campaign import config as test_module
from blueetl.utils import dump_yaml, load_yaml
from tests.unit.utils import TEST_DATA_PATH, assert_frame_equal


@pytest.mark.parametrize(
    "input_dict, expected_obj",
    [
        (
            "xarray_config_dict_simple",
            "blueetl_config_obj_simple",
        ),
        (
            "xarray_config_dict_coupled",
            "blueetl_config_obj_coupled",
        ),
        (
            "blueetl_config_dict_simple",
            "blueetl_config_obj_simple",
        ),
        (
            "blueetl_config_dict_coupled",
            "blueetl_config_obj_coupled",
        ),
    ],
)
def test_simulations_config_load(tmp_path, input_dict, expected_obj, lazy_fixture):
    input_dict = lazy_fixture(input_dict)
    expected_obj = lazy_fixture(expected_obj)
    config_path = tmp_path / "config.yaml"
    dump_yaml(config_path, input_dict)

    result = test_module.SimulationCampaign.load(config_path)

    assert isinstance(result, test_module.SimulationCampaign)
    assert result.name == expected_obj.name
    assert result.attrs == expected_obj.attrs
    assert_frame_equal(result._data, expected_obj._data)
    assert result == expected_obj


@pytest.mark.parametrize(
    "input_obj, expected_dict",
    [
        (
            "blueetl_config_obj_simple",
            "blueetl_config_dict_simple",
        ),
        (
            "blueetl_config_obj_coupled",
            "blueetl_config_dict_coupled",
        ),
    ],
)
def test_simulations_config_dump(tmp_path, input_obj, expected_dict, lazy_fixture):
    input_obj = lazy_fixture(input_obj)
    expected_dict = lazy_fixture(expected_dict)
    config_path = tmp_path / "config.yaml"

    input_obj.dump(config_path)

    content = load_yaml(config_path)
    assert content == expected_dict


@pytest.mark.parametrize(
    "input_dict",
    [
        "blueetl_config_dict_simple",
        "blueetl_config_dict_coupled",
    ],
)
def test_simulations_config_load_dump_roundtrip(tmp_path, input_dict, lazy_fixture):
    input_dict = lazy_fixture(input_dict)
    config_path_1 = tmp_path / "config_1.yaml"
    config_path_2 = tmp_path / "config_2.yaml"
    dump_yaml(config_path_1, input_dict)
    simulations_config_1 = test_module.SimulationCampaign.load(config_path_1)
    simulations_config_1.dump(config_path_2)
    simulations_config_2 = test_module.SimulationCampaign.load(config_path_2)

    assert simulations_config_1.name == simulations_config_2.name
    assert simulations_config_1.attrs == simulations_config_2.attrs
    assert_frame_equal(simulations_config_1._data, simulations_config_2._data)
    assert simulations_config_1 == simulations_config_2


@pytest.mark.parametrize(
    "input_dict, expected_obj",
    [
        (
            "blueetl_config_dict_simple",
            "blueetl_config_obj_simple",
        ),
        (
            "blueetl_config_dict_coupled",
            "blueetl_config_obj_coupled",
        ),
    ],
)
def test_simulations_config_from_dict(input_dict, expected_obj, lazy_fixture):
    input_dict = lazy_fixture(input_dict)
    expected_obj = lazy_fixture(expected_obj)
    result = test_module.SimulationCampaign.from_dict(input_dict)

    assert isinstance(result, test_module.SimulationCampaign)
    assert result == expected_obj


@pytest.mark.parametrize(
    "input_dict, expected_obj",
    [
        (
            "xarray_config_dict_simple",
            "blueetl_config_obj_simple",
        ),
        (
            "xarray_config_dict_coupled",
            "blueetl_config_obj_coupled",
        ),
    ],
)
def test_simulations_config_from_xarray_dict(input_dict, expected_obj, lazy_fixture):
    input_dict = lazy_fixture(input_dict)
    expected_obj = lazy_fixture(expected_obj)
    result = test_module.SimulationCampaign.from_xarray_dict(input_dict)

    assert isinstance(result, test_module.SimulationCampaign)
    assert result == expected_obj


@pytest.mark.parametrize(
    "input_obj, expected_dict",
    [
        (
            "blueetl_config_obj_simple",
            "blueetl_config_dict_simple",
        ),
        (
            "blueetl_config_obj_coupled",
            "blueetl_config_dict_coupled",
        ),
    ],
)
def test_simulations_config_to_dict(input_obj, expected_dict, lazy_fixture):
    input_obj = lazy_fixture(input_obj)
    expected_dict = lazy_fixture(expected_dict)

    result = input_obj.to_dict()

    assert isinstance(result, dict)
    assert result == expected_dict


@pytest.mark.parametrize(
    "input_obj, expected_dict",
    [
        (
            "blueetl_config_obj_simple",
            "xarray_config_dict_simple",
        ),
        (
            "blueetl_config_obj_coupled",
            "xarray_config_dict_coupled",
        ),
    ],
)
def test_simulations_config_to_xarray_dict(input_obj, expected_dict, lazy_fixture):
    input_obj = lazy_fixture(input_obj)
    expected_dict = lazy_fixture(expected_dict)

    result = input_obj.to_xarray_dict()

    assert isinstance(result, dict)
    result = json.loads(json.dumps(result))  # convert tuples to lists
    assert result == expected_dict


@pytest.mark.parametrize(
    "input_obj, expected_df",
    [
        (
            "blueetl_config_obj_simple",
            "blueetl_config_dataframe_simple",
        ),
        (
            "blueetl_config_obj_coupled",
            "blueetl_config_dataframe_coupled",
        ),
    ],
)
def test_simulations_config_get_all(input_obj, expected_df, lazy_fixture):
    input_obj = lazy_fixture(input_obj)
    expected_df = lazy_fixture(expected_df)

    result = input_obj.get()

    assert isinstance(result, pd.DataFrame)
    assert_frame_equal(result, expected_df)
    assert result is not input_obj._data


@pytest.mark.parametrize(
    "input_obj",
    [
        "blueetl_config_obj_simple",
        "blueetl_config_obj_coupled",
    ],
)
@pytest.mark.parametrize(
    "filename, expected",
    [
        ("circuit_sonata.json", True),
        ("CircuitConfig", False),
    ],
)
def test_simulations_config_is_sonata(input_obj, filename, expected, lazy_fixture):
    input_obj = lazy_fixture(input_obj)
    input_obj.attrs["circuit_config"] = f"/path/to/{filename}"
    result = input_obj.is_sonata()
    assert result == expected


@pytest.mark.parametrize(
    "input_obj",
    [
        "blueetl_config_obj_simple",
        "blueetl_config_obj_coupled",
    ],
)
def test_simulations_config_is_sonata_raises(input_obj, lazy_fixture):
    input_obj = lazy_fixture(input_obj)
    del input_obj.attrs["circuit_config"]
    with pytest.raises(RuntimeError, match="circuit_config is missing in the simulation campaign"):
        input_obj.is_sonata()


@pytest.mark.parametrize(
    "input_obj, expected_len",
    [
        ("blueetl_config_obj_simple", 4),
        ("blueetl_config_obj_coupled", 2),
    ],
)
def test_simulations_config_len(input_obj, expected_len, lazy_fixture):
    input_obj = lazy_fixture(input_obj)
    result = len(input_obj)
    assert result == expected_len


@pytest.mark.parametrize(
    "input_obj, expected_len",
    [
        ("blueetl_config_obj_simple", 4),
        ("blueetl_config_obj_coupled", 2),
    ],
)
def test_simulations_config_iter(input_obj, expected_len, lazy_fixture):
    input_obj = lazy_fixture(input_obj)
    count = 0
    for sim in input_obj:
        assert isinstance(sim, test_module.SimulationRow)
        count += 1
    assert count == expected_len


@pytest.mark.parametrize(
    "input_obj, expected_len",
    [
        ("blueetl_config_obj_simple", 4),
        ("blueetl_config_obj_coupled", 2),
    ],
)
def test_simulations_config_getitem(input_obj, expected_len, lazy_fixture):
    input_obj = lazy_fixture(input_obj)
    for i in range(expected_len):
        sim = input_obj[i]
        assert isinstance(sim, test_module.SimulationRow)


@pytest.mark.parametrize(
    "input_obj, expected_data, expected_index",
    [
        ("blueetl_config_obj_simple", [[2.0, 4.0]], [3]),
        ("blueetl_config_obj_coupled", [[2.0, 4.0]], [1]),
    ],
)
def test_simulations_config_get(input_obj, expected_data, expected_index, lazy_fixture):
    input_obj = lazy_fixture(input_obj)

    result = input_obj.get(ca=2.0, depolarization=4.0)

    columns = ["ca", "depolarization"]
    expected_df = pd.DataFrame(expected_data, columns=columns, index=expected_index)
    assert_frame_equal(result[columns], expected_df)


@pytest.mark.parametrize(
    "input_obj, expected_ids",
    [
        ("blueetl_config_obj_simple", [3]),
        ("blueetl_config_obj_coupled", [1]),
    ],
)
def test_simulations_config_ids(input_obj, expected_ids, lazy_fixture):
    input_obj = lazy_fixture(input_obj)

    result = input_obj.ids(ca=2.0, depolarization=4.0)

    assert_array_equal(result, expected_ids)


@pytest.mark.parametrize(
    "config_file",
    [
        "config_01_relative_bbp_workflow.json",
        "config_02_relative_blueetl.yaml",
    ],
)
def test_simulations_config_with_relative_paths(config_file):
    config_file = TEST_DATA_PATH / "simulation_campaign" / config_file

    result = test_module.SimulationCampaign.load(config_file)

    assert Path(result.attrs["path_prefix"]).is_absolute()
    assert Path(result.attrs["circuit_config"]).is_absolute()



from unittest.mock import MagicMock, Mock, patch

import pandas as pd
import pytest
from pandas.testing import assert_frame_equal

from blueetl.campaign.config import SimulationCampaign
from blueetl.extract import simulations as test_module
from blueetl.utils import ensure_dtypes


def _get_mock_simulation(n=0, exists=True, is_complete=True):
    mock_circuit = Mock()
    mock_circuit.checksum.return_value = n
    mock_simulation = Mock()
    mock_simulation.circuit = mock_circuit
    mock_simulation.exists.return_value = exists
    mock_simulation.is_complete.return_value = is_complete
    return mock_simulation


@patch(f"{test_module.__name__}.Simulation", autospec=True)
def test_simulations_from_config(mock_simulation_class):
    mock_simulation0 = _get_mock_simulation()
    mock_simulation1 = _get_mock_simulation()
    mock_circuit0 = mock_simulation0.circuit
    mock_circuit1 = mock_simulation1.circuit
    mock_simulation_class.from_file.side_effect = [mock_simulation0, mock_simulation1]
    config = MagicMock(SimulationCampaign)
    config.get.return_value = pd.DataFrame(
        [
            {"simulation_path": "path1", "seed": 10, "Grating Orientation (degrees)": 0},
            {"simulation_path": "path2", "seed": 11, "Grating Orientation (degrees)": 45},
        ]
    )
    result = test_module.Simulations.from_config(config)

    expected_df = pd.DataFrame(
        [
            {
                "simulation_path": "path1",
                "seed": 10,
                "Grating Orientation (degrees)": 0,
                "simulation_id": 0,
                "circuit_id": 0,
                "simulation": mock_simulation0,
                "circuit": mock_circuit0,
            },
            {
                "simulation_path": "path2",
                "seed": 11,
                "Grating Orientation (degrees)": 45,
                "simulation_id": 1,
                "circuit_id": 0,
                "simulation": mock_simulation1,
                "circuit": mock_circuit0,
            },
        ]
    )
    expected_df = ensure_dtypes(expected_df)
    assert isinstance(result, test_module.Simulations)
    assert_frame_equal(result.df, expected_df)

    assert config.get.call_count == 1
    assert mock_simulation_class.from_file.call_count == 2
    assert mock_circuit0 != mock_circuit1
    assert mock_simulation0.exists.call_count == 1
    assert mock_simulation1.exists.call_count == 1
    assert mock_simulation0.is_complete.call_count == 1
    assert mock_simulation1.is_complete.call_count == 1


@patch(f"{test_module.__name__}.Simulation", autospec=True)
def test_simulations_from_config_filtered_by_simulation_id(mock_simulation_class):
    mock_simulation0 = _get_mock_simulation()
    mock_simulation1 = _get_mock_simulation()
    mock_circuit0 = mock_simulation0.circuit
    mock_circuit1 = mock_simulation1.circuit
    mock_simulation_class.from_file.side_effect = [mock_simulation0, mock_simulation1]
    config = MagicMock(SimulationCampaign)
    config.get.return_value = pd.DataFrame(
        [
            {"simulation_path": "path1", "seed": 10, "Grating Orientation (degrees)": 0},
            {"simulation_path": "path2", "seed": 11, "Grating Orientation (degrees)": 45},
        ]
    )
    result = test_module.Simulations.from_config(config, query={"simulation_id": [1]})

    expected_df = pd.DataFrame(
        [
            {
                "simulation_path": "path2",
                "seed": 11,
                "Grating Orientation (degrees)": 45,
                "simulation_id": 1,
                "circuit_id": 0,
                "simulation": mock_simulation1,
                "circuit": mock_circuit0,
            },
        ],
    )
    expected_df = ensure_dtypes(expected_df)
    assert isinstance(result, test_module.Simulations)
    assert_frame_equal(result.df, expected_df)

    assert config.get.call_count == 1
    assert mock_simulation_class.from_file.call_count == 2
    assert mock_circuit0 != mock_circuit1
    assert mock_simulation0.exists.call_count == 1
    assert mock_simulation1.exists.call_count == 1
    assert mock_simulation0.is_complete.call_count == 1
    assert mock_simulation1.is_complete.call_count == 1


@patch(f"{test_module.__name__}.Simulation", autospec=True)
def test_simulations_from_config_without_spikes(mock_simulation_class):
    mock_simulation0 = _get_mock_simulation(is_complete=False)
    mock_simulation1 = _get_mock_simulation(is_complete=True)
    mock_circuit0 = mock_simulation0.circuit
    mock_circuit1 = mock_simulation1.circuit
    mock_simulation_class.from_file.side_effect = [mock_simulation0, mock_simulation1]
    config = MagicMock(SimulationCampaign)
    config.get.return_value = pd.DataFrame(
        [
            {"simulation_path": "path1", "seed": 10, "Grating Orientation (degrees)": 0},
            {"simulation_path": "path2", "seed": 11, "Grating Orientation (degrees)": 45},
        ]
    )

    result = test_module.Simulations.from_config(config)

    expected_df = pd.DataFrame(
        [
            {
                "simulation_path": "path2",
                "seed": 11,
                "Grating Orientation (degrees)": 45,
                "simulation_id": 1,
                "circuit_id": 0,
                "simulation": mock_simulation1,
                "circuit": mock_circuit0,
            },
        ],
    )
    expected_df = ensure_dtypes(expected_df)
    assert isinstance(result, test_module.Simulations)
    assert_frame_equal(result.df, expected_df)

    assert config.get.call_count == 1
    assert mock_simulation_class.from_file.call_count == 2
    assert mock_circuit0 != mock_circuit1
    assert mock_simulation0.exists.call_count == 1
    assert mock_simulation1.exists.call_count == 1
    assert mock_simulation0.is_complete.call_count == 1
    assert mock_simulation1.is_complete.call_count == 1


@patch(f"{test_module.__name__}.Simulation", autospec=True)
def test_simulations_from_config_first_nonexistent(mock_simulation_class):
    mock_simulation0 = _get_mock_simulation(exists=False)
    mock_simulation1 = _get_mock_simulation(exists=True)
    mock_circuit0 = mock_simulation0.circuit
    mock_circuit1 = mock_simulation1.circuit
    mock_simulation_class.from_file.side_effect = [mock_simulation0, mock_simulation1]
    config = MagicMock(SimulationCampaign)
    config.get.return_value = pd.DataFrame(
        [
            {"simulation_path": "path1", "seed": 10, "Grating Orientation (degrees)": 0},
            {"simulation_path": "path2", "seed": 11, "Grating Orientation (degrees)": 45},
        ]
    )

    result = test_module.Simulations.from_config(config)

    expected_df = pd.DataFrame(
        [
            # circuit_id is 1 and not 0, and circuit is mock_circuit1 and not 0,
            # because the first simulation and circuit cannot be loaded and hashed
            {
                "simulation_path": "path2",
                "seed": 11,
                "Grating Orientation (degrees)": 45,
                "simulation_id": 1,
                "circuit_id": 1,
                "simulation": mock_simulation1,
                "circuit": mock_circuit1,
            },
        ],
    )
    expected_df = ensure_dtypes(expected_df)
    assert isinstance(result, test_module.Simulations)
    assert_frame_equal(result.df, expected_df)

    assert config.get.call_count == 1
    assert mock_simulation_class.from_file.call_count == 2
    assert mock_circuit0 != mock_circuit1
    assert mock_simulation0.exists.call_count == 1
    assert mock_simulation1.exists.call_count == 1
    assert mock_simulation0.is_complete.call_count == 0
    assert mock_simulation1.is_complete.call_count == 1


@patch(f"{test_module.__name__}.Simulation", autospec=True)
def test_simulations_from_pandas_load_complete_campaign(mock_simulation_class):
    mock_simulation0 = _get_mock_simulation()
    mock_simulation1 = _get_mock_simulation()
    mock_circuit0 = mock_simulation0.circuit
    mock_circuit1 = mock_simulation1.circuit
    mock_simulation_class.from_file.side_effect = [mock_simulation0, mock_simulation1]
    df = pd.DataFrame(
        [
            {
                "simulation_path": "path1",
                "seed": 10,
                "Grating Orientation (degrees)": 0,
                "simulation_id": 0,
                "circuit_id": 0,
            },
            {
                "simulation_path": "path2",
                "seed": 11,
                "Grating Orientation (degrees)": 45,
                "simulation_id": 1,
                "circuit_id": 0,
            },
        ]
    )
    result = test_module.Simulations.from_pandas(df)
    expected_df = pd.DataFrame(
        [
            {
                "simulation_path": "path1",
                "seed": 10,
                "Grating Orientation (degrees)": 0,
                "simulation_id": 0,
                "circuit_id": 0,
                "simulation": mock_simulation0,
                "circuit": mock_circuit0,
            },
            {
                "simulation_path": "path2",
                "seed": 11,
                "Grating Orientation (degrees)": 45,
                "simulation_id": 1,
                "circuit_id": 0,
                "simulation": mock_simulation1,
                "circuit": mock_circuit0,
            },
        ]
    )
    expected_df = ensure_dtypes(expected_df)
    assert isinstance(result, test_module.Simulations)
    assert_frame_equal(result.df, expected_df)

    assert mock_simulation_class.from_file.call_count == 2
    assert mock_circuit0 != mock_circuit1
    assert mock_simulation0.exists.call_count == 1
    assert mock_simulation1.exists.call_count == 1
    assert mock_simulation0.is_complete.call_count == 1
    assert mock_simulation1.is_complete.call_count == 1


@patch(f"{test_module.__name__}.Simulation", autospec=True)
def test_simulations_from_pandas_load_incomplete_campaign(mock_simulation_class):
    mock_simulation0 = _get_mock_simulation()
    mock_simulation1 = _get_mock_simulation()
    mock_circuit0 = mock_simulation0.circuit
    mock_circuit1 = mock_simulation1.circuit
    mock_simulation_class.from_file.side_effect = [mock_simulation0, mock_simulation1]
    df = pd.DataFrame(
        [
            {
                "simulation_path": "path1",
                "seed": 10,
                "Grating Orientation (degrees)": 0,
                "simulation_id": 10,
                "circuit_id": 5,
            },
            {
                "simulation_path": "path2",
                "seed": 11,
                "Grating Orientation (degrees)": 45,
                "simulation_id": 11,
                "circuit_id": 5,
            },
        ]
    )
    result = test_module.Simulations.from_pandas(df)
    expected_df = pd.DataFrame(
        [
            {
                "simulation_path": "path1",
                "seed": 10,
                "Grating Orientation (degrees)": 0,
                "simulation_id": 10,
                "circuit_id": 5,
                "simulation": mock_simulation0,
                "circuit": mock_circuit0,
            },
            {
                "simulation_path": "path2",
                "seed": 11,
                "Grating Orientation (degrees)": 45,
                "simulation_id": 11,
                "circuit_id": 5,
                "simulation": mock_simulation1,
                "circuit": mock_circuit0,
            },
        ]
    )
    expected_df = ensure_dtypes(expected_df)
    assert isinstance(result, test_module.Simulations)
    assert_frame_equal(result.df, expected_df)

    assert mock_simulation_class.from_file.call_count == 2
    assert mock_circuit0 != mock_circuit1
    assert mock_simulation0.exists.call_count == 1
    assert mock_simulation1.exists.call_count == 1
    assert mock_simulation0.is_complete.call_count == 1
    assert mock_simulation1.is_complete.call_count == 1


@patch(f"{test_module.__name__}.Simulation", autospec=True)
def test_simulations_from_pandas_load_inconsistent_campaign(mock_simulation_class):
    # the circuits of the mocked simulations have different circuit_hash,
    # while the circuit_ids in the dataframe are the same
    mock_simulation0 = _get_mock_simulation(0)
    mock_simulation1 = _get_mock_simulation(1)
    mock_circuit0 = mock_simulation0.circuit
    mock_circuit1 = mock_simulation1.circuit
    mock_simulation_class.from_file.side_effect = [mock_simulation0, mock_simulation1]
    df = pd.DataFrame(
        [
            {
                "simulation_path": "path1",
                "seed": 10,
                "Grating Orientation (degrees)": 0,
                "simulation_id": 0,
                "circuit_id": 0,
            },
            {
                "simulation_path": "path2",
                "seed": 11,
                "Grating Orientation (degrees)": 45,
                "simulation_id": 1,
                "circuit_id": 0,
            },
        ]
    )
    with pytest.raises(test_module.InconsistentSimulations, match="Inconsistent hash and id"):
        test_module.Simulations.from_pandas(df)

    assert mock_simulation_class.from_file.call_count == 2
    assert mock_circuit0 != mock_circuit1
    assert mock_simulation0.exists.call_count == 1
    assert mock_simulation1.exists.call_count == 1
    assert mock_simulation0.is_complete.call_count == 1
    assert mock_simulation1.is_complete.call_count == 0


@patch(f"{test_module.__name__}.Simulation", autospec=True)
def test_simulations_from_pandas_first_nonexistent(mock_simulation_class):
    mock_simulation0 = _get_mock_simulation(exists=False)
    mock_simulation1 = _get_mock_simulation(exists=True)
    mock_circuit0 = mock_simulation0.circuit
    mock_circuit1 = mock_simulation1.circuit
    mock_simulation_class.from_file.side_effect = [mock_simulation0, mock_simulation1]
    df = pd.DataFrame(
        [
            {
                "simulation_path": "path1",
                "seed": 10,
                "Grating Orientation (degrees)": 0,
                "simulation_id": 0,
                "circuit_id": 0,
            },
            {
                "simulation_path": "path2",
                "seed": 11,
                "Grating Orientation (degrees)": 45,
                "simulation_id": 1,
                "circuit_id": 0,
            },
        ]
    )
    with pytest.raises(test_module.InconsistentSimulations, match="Inconsistent cache"):
        test_module.Simulations.from_pandas(df)

    assert mock_simulation_class.from_file.call_count == 2
    assert mock_circuit0 != mock_circuit1
    assert mock_simulation0.exists.call_count == 1
    assert mock_simulation1.exists.call_count == 1
    assert mock_simulation0.is_complete.call_count == 0
    assert mock_simulation1.is_complete.call_count == 1


@patch(f"{test_module.__name__}.Simulation", autospec=True)
def test_simulations_from_pandas_filtered_by_simulation_id(mock_simulation_class):
    mock_simulation0 = _get_mock_simulation()
    mock_simulation1 = _get_mock_simulation()
    mock_circuit0 = mock_simulation0.circuit
    mock_circuit1 = mock_simulation1.circuit
    mock_simulation_class.from_file.side_effect = [mock_simulation0, mock_simulation1]

    df = pd.DataFrame(
        [
            {
                "simulation_path": "path1",
                "seed": 10,
                "Grating Orientation (degrees)": 0,
                "simulation_id": 0,
                "circuit_id": 0,
            },
            {
                "simulation_path": "path2",
                "seed": 11,
                "Grating Orientation (degrees)": 45,
                "simulation_id": 1,
                "circuit_id": 0,
            },
        ]
    )
    result = test_module.Simulations.from_pandas(df, query={"simulation_id": [1]})
    expected_df = pd.DataFrame(
        [
            {
                "simulation_path": "path2",
                "seed": 11,
                "Grating Orientation (degrees)": 45,
                "simulation_id": 1,
                "circuit_id": 0,
                "simulation": mock_simulation1,
                "circuit": mock_circuit0,
            },
        ],
    )
    expected_df = ensure_dtypes(expected_df)
    assert isinstance(result, test_module.Simulations)

    assert_frame_equal(result.df, expected_df)

    assert mock_simulation_class.from_file.call_count == 2
    assert mock_circuit0 != mock_circuit1
    assert mock_simulation0.exists.call_count == 1
    assert mock_simulation1.exists.call_count == 1
    assert mock_simulation0.is_complete.call_count == 1
    assert mock_simulation1.is_complete.call_count == 1


def test_simulations_to_pandas():
    mock_simulation0 = _get_mock_simulation()
    mock_simulation1 = _get_mock_simulation()
    mock_circuit0 = mock_simulation0.circuit
    mock_circuit1 = mock_simulation1.circuit
    df = pd.DataFrame(
        [
            {
                "simulation_path": "path1",
                "seed": 10,
                "Grating Orientation (degrees)": 0,
                "simulation_id": 0,
                "circuit_id": 0,
                "simulation": mock_simulation0,
                "circuit": mock_circuit0,
            },
            {
                "simulation_path": "path2",
                "seed": 11,
                "Grating Orientation (degrees)": 45,
                "simulation_id": 1,
                "circuit_id": 0,
                "simulation": mock_simulation1,
                "circuit": mock_circuit1,
            },
        ]
    )
    result_df = test_module.Simulations(df, cached=False, filtered=False).to_pandas()
    expected_df = pd.DataFrame(
        [
            {
                "simulation_path": "path1",
                "seed": 10,
                "Grating Orientation (degrees)": 0,
                "simulation_id": 0,
                "circuit_id": 0,
            },
            {
                "simulation_path": "path2",
                "seed": 11,
                "Grating Orientation (degrees)": 45,
                "simulation_id": 1,
                "circuit_id": 0,
            },
        ]
    )
    expected_df = ensure_dtypes(expected_df)
    assert_frame_equal(result_df, expected_df)
    assert mock_circuit0 != mock_circuit1



"""Common utils from BlueCV."""

from elephant.conversion import BinnedSpikeTrain
from neo import SpikeTrain
from quantities import ms


def to_spiketrains(arrays, t_start, t_stop):
    """Convert arrays to spiketrains."""
    return [SpikeTrain(spiketrain * ms, t_start=t_start, t_stop=t_stop) for spiketrain in arrays]


def to_binned_spiketrain(ST, bin_size=5 * ms):
    """Convert a spiketrain to a binned spiketrain."""
    return BinnedSpikeTrain(ST, bin_size=bin_size)



"""Bluepysnap NodeSets implementation."""

from bluepysnap.node_sets import NodeSets

from blueetl.adapters.interfaces.node_sets import NodeSetsInterface


class NodeSetsImpl(NodeSetsInterface[NodeSets]):
    """Bluepysnap NodeSets implementation."""

    def update(self, other: NodeSetsInterface) -> None:
        """Update the wrapped node sets."""
        self._node_sets.update(other.instance)



"""Interfaces for Simulation."""

from abc import ABC, abstractmethod
from collections.abc import Mapping
from typing import Generic, Optional, TypeVar

import pandas as pd

from blueetl.adapters.interfaces.circuit import CircuitInterface
from blueetl.utils import CachedPropertyMixIn

SimulationT = TypeVar("SimulationT")


class PopulationSpikesReportInterface(ABC):
    """PopulationSpikesReport Interface."""

    @abstractmethod
    def get(self, group=None, t_start=None, t_stop=None) -> pd.Series:
        """Return the spikes report for the specified group and interval."""


class PopulationReportInterface(ABC):
    """PopulationReport Interface."""

    @abstractmethod
    def get(self, group=None, t_start=None, t_stop=None, t_step=None) -> pd.DataFrame:
        """Return the report for the specified group and interval."""


class SimulationInterface(Generic[SimulationT], CachedPropertyMixIn, ABC):
    """Simulation Interface."""

    def __init__(self, simulation: SimulationT) -> None:
        """Init the simulation interface with the specified simulation."""
        self._simulation = simulation

    @property
    def instance(self) -> SimulationT:
        """Return the wrapped instance."""
        return self._simulation

    @abstractmethod
    def is_complete(self) -> bool:
        """Return True if the simulation is complete, False otherwise."""

    @property
    @abstractmethod
    def circuit(self) -> CircuitInterface:
        """Return the circuit used for the simulation."""

    @property
    @abstractmethod
    def spikes(self) -> Mapping[Optional[str], PopulationSpikesReportInterface]:
        """Return the spikes report as a dict: population -> report."""

    @property
    @abstractmethod
    def reports(self) -> Mapping[str, Mapping[Optional[str], PopulationReportInterface]]:
        """Return the reports as a dict: name -> population -> report."""



from unittest.mock import Mock

import pytest

from blueetl import resolver as test_module


def test_object_resolver():
    obj = Mock()
    obj.a.b.c = 123
    resolver = test_module.AttrResolver(obj)

    result = resolver.get("a.b.c")

    assert result == 123


def test_object_resolver_error():
    obj = Mock()
    obj.a.b = 123
    resolver = test_module.AttrResolver(obj)

    with pytest.raises(test_module.ResolverError, match="Impossible to resolve a.b.c at level c"):
        resolver.get("a.b.c")


def test_dict_resolver():
    obj = {"a": {"b": {"c": 123}}}
    resolver = test_module.ItemResolver(obj)

    result = resolver.get("a.b.c")

    assert result == 123


def test_dict_resolver_error():
    obj = {"a": {"b": 123}}
    resolver = test_module.ItemResolver(obj)

    with pytest.raises(test_module.ResolverError, match="Impossible to resolve a.b.c at level c"):
        resolver.get("a.b.c")
